import json
from typing import Dict, Iterable

import numpy as np
from clarifai_grpc.grpc.api import resources_pb2

from clarifai.runners.utils import data_types


class Serializer:

  def serialize(self, data_proto, value):
    pass

  def deserialize(self, data_proto):
    pass

  def handles_list(self):
    return False


def is_repeated_field(field_name):
  descriptor = resources_pb2.Data.DESCRIPTOR.fields_by_name.get(field_name)
  return descriptor and descriptor.label == descriptor.LABEL_REPEATED


class AtomicFieldSerializer(Serializer):

  def __init__(self, field_name):
    self.field_name = field_name

  def serialize(self, data_proto, value):
    try:
      setattr(data_proto, self.field_name, value)
    except TypeError as e:
      raise TypeError(f"Incompatible type for {self.field_name}: {type(value)}") from e

  def deserialize(self, data_proto):
    return getattr(data_proto, self.field_name)


class MessageSerializer(Serializer):

  def __init__(self, field_name, message_class):
    self.field_name = field_name
    self.message_class = message_class
    self.is_repeated_field = is_repeated_field(field_name)

  def handles_list(self):
    return self.is_repeated_field

  def serialize(self, data_proto, value):
    value = self.message_class.from_value(value).to_proto()
    dst = getattr(data_proto, self.field_name)
    try:
      if self.is_repeated_field:
        dst.add().CopyFrom(value)
      else:
        dst.CopyFrom(value)
    except TypeError as e:
      raise TypeError(f"Incompatible type for {self.field_name}: {type(value)}") from e

  def serialize_list(self, data_proto, values):
    assert self.is_repeated_field
    dst = getattr(data_proto, self.field_name)
    dst.extend([self.message_class.from_value(value).to_proto() for value in values])

  def deserialize(self, data_proto):
    src = getattr(data_proto, self.field_name)
    if self.is_repeated_field:
      values = [self.message_class.from_proto(x) for x in src]
      if len(values) == 1:
        return values[0]
      return values if values else None
    else:
      if not data_proto.HasField(self.field_name):
        return None
      return self.message_class.from_proto(src)

  def deserialize_list(self, data_proto):
    assert self.is_repeated_field
    src = getattr(data_proto, self.field_name)
    return [self.message_class.from_proto(x) for x in src]


class NDArraySerializer(Serializer):

  def __init__(self, field_name, as_list=False):
    self.field_name = field_name
    self.as_list = as_list

  def serialize(self, data_proto, value):
    if self.as_list and not isinstance(value, Iterable):
      raise TypeError(f"Expected list, got {type(value)}")
    value = np.asarray(value)
    if not np.issubdtype(value.dtype, np.number):
      raise TypeError(f"Expected number array, got {value.dtype}")
    proto = getattr(data_proto, self.field_name)
    proto.buffer = value.tobytes()
    proto.shape.extend(value.shape)
    proto.dtype = str(value.dtype)

  def deserialize(self, data_proto):
    proto = getattr(data_proto, self.field_name)
    if not proto.buffer:
      return None
    array = np.frombuffer(proto.buffer, dtype=np.dtype(proto.dtype)).reshape(proto.shape)
    if self.as_list:
      return array.tolist()
    return array


class JSONSerializer(Serializer):

  def __init__(self, field_name, type=None):
    self.field_name = field_name
    self.type = type

  def serialize(self, data_proto, value):
    #if self.type is not None and not isinstance(value, self.type):
    #  raise TypeError(f"Expected {self.type}, got {type(value)}")
    try:
      setattr(data_proto, self.field_name, json.dumps(value))
    except TypeError as e:
      raise TypeError(f"Incompatible type for {self.field_name}: {type(value)}") from e

  def deserialize(self, data_proto):
    value = getattr(data_proto, self.field_name)
    if not value:
      return None
    return json.loads(value)


class ListSerializer(Serializer):

  def __init__(self, inner_serializer):
    self.field_name = 'parts'
    self.inner_serializer = inner_serializer

  def handles_list(self):
    # if handles_list() is called on this serializer, it means that we're
    # trying to serialize a list of lists. In this case, we need to use
    # parts[] for the outer list, so we return False here (we can't inline it).
    return False

  def serialize(self, data_proto, value):
    if not isinstance(value, Iterable):
      raise TypeError(f"Expected iterable, got {type(value)}")
    if self.inner_serializer.handles_list():
      self.inner_serializer.serialize_list(data_proto, value)
    else:
      for item in value:
        part = data_proto.parts.add()
        self.inner_serializer.serialize(part.data, item)

  def deserialize(self, data_proto):
    if self.inner_serializer.handles_list():
      return self.inner_serializer.deserialize_list(data_proto)
    return [self.inner_serializer.deserialize(part.data) for part in data_proto.parts]


class TupleSerializer(Serializer):

  def __init__(self, inner_serializers):
    self.field_name = 'parts'
    self.inner_serializers = inner_serializers

  def serialize(self, data_proto, value):
    if not isinstance(value, (tuple, list)):
      raise TypeError(f"Expected tuple, got {type(value)}")
    if len(value) != len(self.inner_serializers):
      raise ValueError(f"Expected tuple of length {len(self.inner_serializers)}, got {len(value)}")
    for i, (serializer, item) in enumerate(zip(self.inner_serializers, value)):
      part = data_proto.parts.add()
      part.id = str(i)
      serializer.serialize(part.data, item)

  def deserialize(self, data_proto):
    if not data_proto.parts and self.inner_serializers:
      return None
    if len(data_proto.parts) != len(self.inner_serializers):
      raise ValueError(
          f"Expected tuple of length {len(self.inner_serializers)}, got {len(data_proto.parts)}")
    return tuple(
        serializer.deserialize(part.data)
        for serializer, part in zip(self.inner_serializers, data_proto.parts))


class NamedFieldsSerializer(Serializer):

  def __init__(self, named_field_serializers: Dict[str, Serializer]):
    self.field_name = 'parts'
    self.named_field_serializers = named_field_serializers

  def serialize(self, data_proto, value):
    for name, serializer in self.named_field_serializers.items():
      if name not in value:
        raise TypeError(f"Missing field {name}")
      part = self._get_part(data_proto, name, add=True)
      serializer.serialize(part.data, value[name])

  def deserialize(self, data_proto):
    if not data_proto.parts and self.named_field_serializers:
      return None
    value = data_types.NamedFields()
    for name, serializer in self.named_field_serializers.items():
      part = self._get_part(data_proto, name)
      value[name] = serializer.deserialize(part.data)
    return value

  def _get_part(self, data_proto, name, add=False):
    for part in data_proto.parts:
      if part.id == name:
        return part
    if add:
      part = data_proto.parts.add()
      part.id = name
      return part
    raise TypeError(f"Missing part with key {name}")


# TODO dict serializer, maybe json only?



import io
import json
from typing import Iterable, List, get_args, get_origin

import numpy as np
from clarifai_grpc.grpc.api.resources_pb2 import Audio as AudioProto
from clarifai_grpc.grpc.api.resources_pb2 import Concept as ConceptProto
from clarifai_grpc.grpc.api.resources_pb2 import Frame as FrameProto
from clarifai_grpc.grpc.api.resources_pb2 import Image as ImageProto
from clarifai_grpc.grpc.api.resources_pb2 import Region as RegionProto
from clarifai_grpc.grpc.api.resources_pb2 import Text as TextProto
from clarifai_grpc.grpc.api.resources_pb2 import Video as VideoProto
from PIL import Image as PILImage


class MessageData:

  def to_proto(self):
    raise NotImplementedError

  @classmethod
  def from_proto(cls, proto):
    raise NotImplementedError

  @classmethod
  def from_value(cls, value):
    if isinstance(value, cls):
      return value
    return cls(value)

  def cast(self, python_type):
    if python_type == self.__class__:
      return self
    raise TypeError(f'Incompatible type for {self.__class__.__name__}: {python_type}')


class NamedFieldsMeta(type):
  """Metaclass to create NamedFields subclasses with __annotations__ when fields are specified."""

  def __call__(cls, *args, **kwargs):
    # Check if keyword arguments are types (used in type annotations)
    if kwargs and all(isinstance(v, type) for v in kwargs.values()):
      # Dynamically create a subclass with __annotations__
      name = f"NamedFields({', '.join(f'{k}:{v.__name__}' for k, v in kwargs.items())})"
      return type(name, (cls,), {'__annotations__': kwargs})
    else:
      # Create a normal instance for runtime data
      return super().__call__(*args, **kwargs)


class NamedFields(metaclass=NamedFieldsMeta):
  """A class that can be used to store named fields with values."""

  def __init__(self, **kwargs):
    for key, value in kwargs.items():
      setattr(self, key, value)

  def items(self):
    return self.__dict__.items()

  def keys(self):
    return self.__dict__.keys()

  def values(self):
    return self.__dict__.values()

  def __contains__(self, key):
    return key in self.__dict__

  def __getitem__(self, key):
    return getattr(self, key)

  def __setitem__(self, key, value):
    setattr(self, key, value)

  def __repr__(self):
    return f"{self.__class__.__name__}({', '.join(f'{key}={value!r}' for key, value in self.__dict__.items())})"

  def __origin__(self):
    return self

  def __args__(self):
    return list(self.keys())


class Stream(Iterable):
  pass


class JSON:

  def __init__(self, value):
    self.value = value

  def __eq__(self, other):
    return self.value == other

  def __bool__(self):
    return bool(self.value)

  def __repr__(self):
    return f"JSON({self.value})"

  def to_json(self):
    return json.dumps(self.value)

  @classmethod
  def from_json(cls, json_str):
    return cls(json.loads(json_str))

  @classmethod
  def from_value(cls, value):
    return cls(value)

  def cast(self, python_type):
    if not isinstance(self.value, python_type):
      raise TypeError(f'Incompatible type {type(self.value)} for {python_type}')
    return self.value


class Text(MessageData):

  def __init__(self, text: str, url: str = None):
    self.text = text
    self.url = url

  def __eq__(self, other):
    if isinstance(other, Text):
      return self.text == other.text and self.url == other.url
    if isinstance(other, str):
      return self.text == other
    return False

  def __bool__(self):
    return bool(self.text) or bool(self.url)

  def __repr__(self) -> str:
    return f"Text(text={self.text!r}, url={self.url!r})"

  def to_proto(self) -> TextProto:
    return TextProto(raw=self.text or '', url=self.url or '')

  @classmethod
  def from_proto(cls, proto: TextProto) -> "Text":
    return cls(proto.raw, proto.url or None)

  @classmethod
  def from_value(cls, value):
    if isinstance(value, str):
      return cls(value)
    if isinstance(value, Text):
      return value
    if isinstance(value, dict):
      return cls(value.get('text'), value.get('url'))
    raise TypeError(f'Incompatible type for Text: {type(value)}')

  def cast(self, python_type):
    if python_type == str:
      return self.text
    if python_type == Text:
      return self
    raise TypeError(f'Incompatible type for Text: {python_type}')


class Concept(MessageData):

  def __init__(self, id: str, name: str, value: float = 1):
    self.id = id
    self.name = name
    self.value = value

  def __repr__(self) -> str:
    return f"Concept(id={self.id!r}, name={self.name!r}, value={self.value})"

  def to_proto(self):
    return ConceptProto(id=self.id, name=self.name, value=self.value)

  @classmethod
  def from_proto(cls, proto: ConceptProto) -> "Concept":
    return cls(proto.id, proto.name, proto.value)


class Region(MessageData):

  def __init__(self,
               proto_region: RegionProto = None,
               box: List[float] = None,
               concepts: List[Concept] = None):
    if proto_region is None:
      proto_region = RegionProto()
    self.proto = proto_region
    # use setters for init vals
    if box and isinstance(
        box, list) and len(box) == 4 and all(isinstance(val, (int, float)) for val in box):
      self.box = box
    if concepts and isinstance(concepts,
                               list) and all(isinstance(concept, Concept) for concept in concepts):
      self.concepts = concepts

  @property
  def box(self) -> List[float]:
    bbox = self.proto.region_info.bounding_box
    return [bbox.left_col, bbox.top_row, bbox.right_col, bbox.bottom_row]  # x1, y1, x2, y2

  @box.setter
  def box(self, value: List[float]):
    bbox = self.proto.region_info.bounding_box
    bbox.left_col, bbox.top_row, bbox.right_col, bbox.bottom_row = value

  @property
  def concepts(self) -> List[Concept]:
    return [Concept.from_proto(proto) for proto in self.proto.data.concepts]

  @concepts.setter
  def concepts(self, value: List[Concept]):
    self.proto.data.concepts.extend([concept.to_proto() for concept in value])

  def __repr__(self) -> str:
    return f"Region(box={self.box}, concepts={self.concepts})"

  def to_proto(self) -> RegionProto:
    return self.proto

  @classmethod
  def from_proto(cls, proto: RegionProto) -> "Region":
    return cls(proto)


class Image(MessageData):

  def __init__(self, proto_image: ImageProto = None, url: str = None, bytes: bytes = None):
    if proto_image is None:
      proto_image = ImageProto()
    self.proto = proto_image
    # use setters for init vals
    if url:
      self.url = url
    if bytes:
      self.bytes = bytes

  @property
  def url(self) -> str:
    return self.proto.url

  @url.setter
  def url(self, value: str):
    self.proto.url = value

  @property
  def bytes(self) -> bytes:
    return self.proto.base64

  @bytes.setter
  def bytes(self, value: bytes):
    self.proto.base64 = value

  def __repr__(self) -> str:
    attrs = []
    if self.url:
      attrs.append(f"url={self.url!r}")
    if self.bytes:
      attrs.append(f"bytes=<{len(self.bytes)} bytes>")
    return f"Image({', '.join(attrs)})"

  @classmethod
  def from_url(cls, url: str) -> "Image":
    proto_image = ImageProto(url=url)
    return cls(proto_image)

  @classmethod
  def from_pil(cls, pil_image: PILImage.Image) -> "Image":
    with io.BytesIO() as output:
      pil_image.save(output, format="PNG")
      image_bytes = output.getvalue()
    proto_image = ImageProto(base64=image_bytes)
    return cls(proto_image)

  def to_pil(self) -> PILImage.Image:
    if not self.proto.base64:
      raise ValueError("Image has no bytes")
    return PILImage.open(io.BytesIO(self.proto.base64))

  def to_numpy(self) -> np.ndarray:
    return np.asarray(self.to_pil())

  def to_proto(self) -> ImageProto:
    return self.proto

  @classmethod
  def from_proto(cls, proto: ImageProto) -> "Image":
    return cls(proto)

  @classmethod
  def from_value(cls, value):
    if isinstance(value, PILImage.Image):
      return cls.from_pil(value)
    if isinstance(value, Image):
      return value
    raise TypeError(f'Incompatible type for Image: {type(value)}')

  def cast(self, python_type):
    if python_type == Image:
      return self
    if python_type in (PILImage.Image, PILImage):
      return self.to_pil()
    if python_type == np.ndarray or get_origin(python_type) == np.ndarray:
      return self.to_numpy()
    raise TypeError(f'Incompatible type for Image: {python_type}')


class Audio(MessageData):

  def __init__(self, proto_audio: AudioProto = None, url: str = None, bytes: bytes = None):
    if proto_audio is None:
      proto_audio = AudioProto()
    self.proto = proto_audio

    if url:
      self.url = url
    if bytes:
      self.bytes = bytes

  @property
  def url(self) -> str:
    return self.proto.url

  @url.setter
  def url(self, value: str):
    self.proto.url = value

  @property
  def bytes(self) -> bytes:
    return self.proto.base64

  @bytes.setter
  def bytes(self, value: bytes):
    self.proto.base64 = value

  @classmethod
  def from_url(cls, url: str) -> "Audio":
    proto_audio = AudioProto(url=url)
    return cls(proto_audio)

  def __repr__(self) -> str:
    attrs = []
    if self.url:
      attrs.append(f"url={self.url!r}")
    if self.bytes:
      attrs.append(f"bytes=<{len(self.bytes)} bytes>")
    return f"Audio({', '.join(attrs)})"

  def to_proto(self) -> AudioProto:
    return self.proto

  @classmethod
  def from_proto(cls, proto: AudioProto) -> "Audio":
    return cls(proto)


class Frame(MessageData):

  def __init__(self,
               proto_frame: FrameProto = None,
               image: Image = None,
               regions: List[Region] = None,
               time: float = None):
    if proto_frame is None:
      proto_frame = FrameProto()
    self.proto = proto_frame
    # use setters for init vals
    if image:
      self.image = image
    if regions:
      self.regions = regions
    if time:
      self.time = time

  @property
  def time(self) -> float:
    # TODO: time is a uint32, so this will overflow at 49.7 days
    # we should be using double or uint64 in the proto instead
    return self.proto.frame_info.time / 1000.0

  @time.setter
  def time(self, value: float):
    self.proto.frame_info.time = int(value * 1000)

  @property
  def image(self) -> Image:
    return Image.from_proto(self.proto.data.image)

  @image.setter
  def image(self, value: Image):
    self.proto.data.image.CopyFrom(value.to_proto())

  @property
  def regions(self) -> List[Region]:
    return [Region(region) for region in self.proto.data.regions]

  @regions.setter
  def regions(self, value: List[Region]):
    self.proto.data.regions.extend([region.proto for region in value])

  def to_proto(self) -> FrameProto:
    return self.proto

  @classmethod
  def from_proto(cls, proto: FrameProto) -> "Frame":
    return cls(proto)


class Video(MessageData):

  def __init__(self, proto_video: VideoProto = None, url: str = None, bytes: bytes = None):
    if proto_video is None:
      proto_video = VideoProto()
    self.proto = proto_video

    if url:
      self.url = url
    if bytes:
      self.bytes = bytes

  @property
  def url(self) -> str:
    return self.proto.url

  @url.setter
  def url(self, value: str):
    self.proto.url = value

  @property
  def bytes(self) -> bytes:
    return self.proto.base64

  @bytes.setter
  def bytes(self, value: bytes):
    self.proto.base64 = value

  @classmethod
  def from_url(cls, url: str) -> "Video":
    proto_video = VideoProto(url=url)
    return cls(proto_video)

  def __repr__(self) -> str:
    attrs = []
    if self.url:
      attrs.append(f"url={self.url!r}")
    if self.bytes:
      attrs.append(f"bytes=<{len(self.bytes)} bytes>")
    return f"Video({', '.join(attrs)})"

  def to_proto(self) -> VideoProto:
    return self.proto

  @classmethod
  def from_proto(cls, proto: VideoProto) -> "Video":
    return cls(proto)


def cast(value, python_type):
  list_type = (get_origin(python_type) == list)
  if isinstance(value, MessageData):
    return value.cast(python_type)
  if list_type and isinstance(value, np.ndarray):
    return value.tolist()
  if list_type and isinstance(value, list):
    if get_args(python_type):
      inner_type = get_args(python_type)[0]
      return [cast(item, inner_type) for item in value]
    if not isinstance(value, Iterable):
      raise TypeError(f'Expected list, got {type(value)}')
  return value



from io import BytesIO

from clarifai_grpc.grpc.api.resources_pb2 import ModelTypeEnumOption
from clarifai_grpc.grpc.api.resources_pb2 import ModelTypeField as InputFieldProto
from clarifai_grpc.grpc.api.resources_pb2 import ModelTypeRangeInfo
from PIL import Image

from clarifai.runners.utils.data_types import MessageData


def image_to_bytes(img: Image.Image, format="JPEG") -> bytes:
  buffered = BytesIO()
  img.save(buffered, format=format)
  img_str = buffered.getvalue()
  return img_str


def bytes_to_image(bytes_img) -> Image.Image:
  img = Image.open(BytesIO(bytes_img))
  return img


def is_openai_chat_format(messages):
  """
    Verify if the given argument follows the OpenAI chat messages format.

    Args:
        messages (list): A list of dictionaries representing chat messages.

    Returns:
        bool: True if valid, False otherwise.
    """
  if not isinstance(messages, list):
    return False

  valid_roles = {"system", "user", "assistant", "function"}

  for msg in messages:
    if not isinstance(msg, dict):
      return False
    if "role" not in msg or "content" not in msg:
      return False
    if msg["role"] not in valid_roles:
      return False

    content = msg["content"]

    # Content should be either a string (text message) or a multimodal list
    if isinstance(content, str):
      continue  # Valid text message

    elif isinstance(content, list):
      for item in content:
        if not isinstance(item, dict):
          return False
  return True


class InputField(MessageData):
  """A field that can be used to store input data."""

  def __init__(
      self,
      default=None,
      description=None,
      min_value=None,
      max_value=None,
      choices=None,
      #  is_param=True
  ):
    self.default = default
    self.description = description
    self.min_value = min_value
    self.max_value = max_value
    self.choices = choices
    # self.is_param = is_param

  def __repr__(self) -> str:
    attrs = []
    if self.default is not None:
      attrs.append(f"default={self.default!r}")
    if self.description is not None:
      attrs.append(f"description={self.description!r}")
    if self.min_value is not None:
      attrs.append(f"min_value={self.min_value!r}")
    if self.max_value is not None:
      attrs.append(f"max_value={self.max_value!r}")
    if self.choices is not None:
      attrs.append(f"choices={self.choices!r}")
    # attrs.append(f"is_param={self.is_param!r}")
    return f"InputField({', '.join(attrs)})"

  def to_proto(self, proto=None) -> InputFieldProto:
    if proto is None:
      proto = InputFieldProto()
    if self.description is not None:
      proto.description = self.description

    if self.choices is not None:
      for choice in self.choices:
        option = ModelTypeEnumOption(id=str(choice))
        proto.model_type_enum_options.append(option)

    proto.required = self.default is None

    if self.min_value is not None or self.max_value is not None:
      range_info = ModelTypeRangeInfo()
      if self.min_value is not None:
        range_info.min = float(self.min_value)
      if self.max_value is not None:
        range_info.max = float(self.max_value)
      proto.model_type_range_info.CopyFrom(range_info)
    # proto.is_param = self.is_param

    if self.default is not None:
      if isinstance(self.default, str) or isinstance(self.default, bool) or isinstance(
          self.default, (int, float)):
        proto.default = str(self.default)
      else:
        import json
        proto.default = json.dumps(self.default)

    return proto

  @classmethod
  def from_proto(cls, proto):
    default = None
    if proto.HasField('default'):
      pb_value = proto.default
      if pb_value.HasField('string_value'):
        default = pb_value.string_value
        try:
          import json
          default = json.loads(default)
        except json.JSONDecodeError:
          pass
      elif pb_value.HasField('number_value'):
        default = pb_value.number_value
        if default.is_integer():
          default = int(default)
        else:
          default = float(default)
      elif pb_value.HasField('bool_value'):
        default = pb_value.bool_value

    choices = [option.id for option in proto.model_type_enum_options
              ] if proto.model_type_enum_options else None

    min_value = None
    max_value = None
    if proto.HasField('model_type_range_info'):
      min_value = proto.model_type_range_info.min
      max_value = proto.model_type_range_info.max
      if min_value.is_integer():
        min_value = int(min_value)
      if max_value.is_integer():
        max_value = int(max_value)

    return cls(
        default=default,
        description=proto.description if proto.description else None,
        min_value=min_value,
        max_value=max_value,
        choices=choices,
        # is_param=proto.is_param
    )

  @classmethod
  def set_default(cls, proto=None, default=None):

    if proto is None:
      proto = InputFieldProto()
    if default is not None:
      if isinstance(default, str) or isinstance(default, bool) or isinstance(
          default, (int, float)):
        proto.default = str(default)
      else:
        import json
        proto.default = json.dumps(default)
    return proto
